import logging
from dataclasses import dataclass
from typing import Union

from ..base_handler import handle_event
from ...core.schedule.model import (
    SchedulingInfo,
    ScheduledJob,
    FixedDelayRepeat,
    DynamicDelayRepeat,
    JobState,
)
from ...core.model.cloud_event import *
from ...core.model.events import *
from ...core.model.trigger_params import TriggerParams
from ...core.model.playbook_hash import playbook_hash
from ...core.schedule.scheduler import Scheduler
from ...utils.decorators import doublewrap
from ...core.active_playbooks import register_playbook, get_function_params_class

SCHEDULED_INTERGATION_TASK = "scheduled_integration_task"
scheduled_callables = {}


@dataclass
class RecurringTriggerEvent(BaseEvent):
    recurrence: int = 0


def __add_callable(func):
    scheduled_callables[func.__name__] = {
        "func": func,
        "action_params": get_function_params_class(func),
    }


@doublewrap
def on_recurring_trigger(func, repeat=1, seconds_delay=None):
    __add_callable(func)
    register_playbook(
        func,
        deploy_on_scheduler_event,
        TriggerParams(repeat=repeat, seconds_delay=seconds_delay),
    )
    return func


@doublewrap
def scheduled_callable(func):
    __add_callable(func)
    return func


class ScheduledIntegrationParams(BaseModel):
    action_func_name: str
    action_params: dict = None
    named_sinks: List[str] = []


def run_scheduled_task(runnable_params: dict, schedule_info: SchedulingInfo):
    scheduled_params = ScheduledIntegrationParams(**runnable_params)

    func_definition = scheduled_callables.get(scheduled_params.action_func_name)
    if not func_definition:
        logging.error(
            f"Scheduled callable cannot be found: {scheduled_params.action_func_name}. Skipping"
        )
        return

    action_params = None
    if func_definition["action_params"]:
        action_params = func_definition["action_params"](
            **scheduled_params.action_params
        )

    trigger_event = RecurringTriggerEvent(recurrence=schedule_info.execution_count)

    return handle_event(
        func_definition["func"],
        trigger_event,
        action_params,
        "scheduler",
        scheduled_params.named_sinks,
    )


Scheduler.register_task(SCHEDULED_INTERGATION_TASK, run_scheduled_task)


def schedule_trigger(
    func,
    playbook_id: str,
    scheduling_params: Union[FixedDelayRepeat, DynamicDelayRepeat],
    named_sinks: List[str],
    action_params=None,
    job_state: JobState = JobState(),
    replace_existing: bool = False,
    standalone_task: bool = False,
):
    integration_params = ScheduledIntegrationParams(
        action_func_name=func.__name__,
        action_params=action_params,
        named_sinks=named_sinks,
    )
    job = ScheduledJob(
        job_id=playbook_id,
        runnable_name=SCHEDULED_INTERGATION_TASK,
        runnable_params=integration_params.dict(),
        state=job_state,
        scheduling_params=scheduling_params,
        replace_existing=replace_existing,
        standalone_task=standalone_task,
    )
    Scheduler.schedule_job(job)


def deploy_on_scheduler_event(
    func, trigger_params: TriggerParams, named_sinks: List[str], action_params=None
):
    if trigger_params.seconds_delay:
        schedule_params = FixedDelayRepeat(
            repeat=trigger_params.repeat, seconds_delay=trigger_params.seconds_delay
        )
    else:
        schedule_params = DynamicDelayRepeat(delay_periods=trigger_params.delays)

    schedule_trigger(
        func=func,
        playbook_id=playbook_hash(func, trigger_params, action_params),
        scheduling_params=schedule_params,
        named_sinks=named_sinks,
        action_params=action_params,
    )



import time
from typing import Type, TypeVar, List, Dict

import hikaru
import json
import yaml
from hikaru.model import *  # *-import is necessary for hikaru subclasses to work
from pydantic import BaseModel

from ...core.model.env_vars import INSTALLATION_NAMESPACE
from .api_client_utils import *
from .templates import get_deployment_yaml

S = TypeVar("S")
T = TypeVar("T")
PYTHON_DEBUGGER_IMAGE = (
    "us-central1-docker.pkg.dev/genuine-flight-317411/devel/python-tools:latest"
)


# TODO: import these from the lookup_pid project
class Process(BaseModel):
    pid: int
    exe: str
    cmdline: List[str]


class ProcessList(BaseModel):
    processes: List[Process]


def get_images(containers: List[Container]) -> Dict[str, str]:
    """
    Takes a list of containers and returns a dict mapping image name to image tag.
    """
    name_to_version = {}
    for container in containers:
        if ":" in container.image:
            image_name, image_tag = container.image.split(":", maxsplit=1)
            name_to_version[image_name] = image_tag
        else:
            name_to_version[container.image] = "<NONE>"
    return name_to_version


def does_daemonset_have_toleration(ds: DaemonSet, toleration_key: str) -> bool:
    return any(t.key == toleration_key for t in ds.spec.template.spec.tolerations)


def does_node_have_taint(node: Node, taint_key: str) -> bool:
    return any(t.key == taint_key for t in node.spec.taints)


class RobustaPod(Pod):
    def exec(self, shell_command: str) -> str:
        """Execute a command inside the pod"""
        return exec_shell_command(
            self.metadata.name, shell_command, self.metadata.namespace
        )

    def get_logs(self, container_name=None, previous=None, tail_lines=None) -> str:
        """
        Fetch pod logs
        """
        if container_name is None:
            container_name = self.spec.containers[0].name
        return get_pod_logs(
            self.metadata.name,
            self.metadata.namespace,
            container_name,
            previous,
            tail_lines,
        )

    @staticmethod
    def create_debugger_pod(
        pod_name: str, node_name: str, debug_image=PYTHON_DEBUGGER_IMAGE, debug_cmd=None
    ) -> "RobustaPod":
        """
        Creates a debugging pod with high privileges
        """
        debugger = RobustaPod(
            apiVersion="v1",
            kind="Pod",
            metadata=ObjectMeta(
                name=to_kubernetes_name(pod_name, "debug-"),
                namespace=INSTALLATION_NAMESPACE,
            ),
            spec=PodSpec(
                hostPID=True,
                nodeName=node_name,
                containers=[
                    Container(
                        name="debugger",
                        image=debug_image,
                        imagePullPolicy="Always",
                        command=prepare_pod_command(debug_cmd),
                        securityContext=SecurityContext(
                            capabilities=Capabilities(add=["SYS_PTRACE", "SYS_ADMIN"]),
                            privileged=True,
                        ),
                    )
                ],
            ),
        )
        # TODO: check the result code
        debugger = debugger.createNamespacedPod(debugger.metadata.namespace).obj
        return debugger

    @staticmethod
    def exec_in_debugger_pod(
        pod_name: str, node_name: str, cmd, debug_image=PYTHON_DEBUGGER_IMAGE
    ) -> str:
        debugger = RobustaPod.create_debugger_pod(pod_name, node_name, debug_image)
        try:
            return debugger.exec(cmd)
        finally:
            RobustaPod.deleteNamespacedPod(
                debugger.metadata.name, debugger.metadata.namespace
            )

    def get_processes(self) -> List[Process]:
        output = RobustaPod.exec_in_debugger_pod(
            self.metadata.name,
            self.spec.nodeName,
            f"/lookup_pid.py {self.metadata.uid}",
        )
        # somehow when doing the exec command the quotes in the json output are converted from " to '
        # we fix this so that we can deserialize the json properly...
        # we should eventually figure out why this is happening
        output = output.replace("'", '"')
        processes = ProcessList(**json.loads(output))
        return processes.processes

    def get_images(self) -> Dict[str, str]:
        return get_images(self.spec.containers)

    def has_direct_owner(self, owner_uid) -> bool:
        for owner in self.metadata.ownerReferences:
            if owner.uid == owner_uid:
                return True
        return False

    @staticmethod
    def find_pods_with_direct_owner(
        namespace: str, owner_uid: str
    ) -> List["RobustaPod"]:
        all_pods: List["RobustaPod"] = PodList.listNamespacedPod(namespace).obj.items
        return list(filter(lambda p: p.has_direct_owner(owner_uid), all_pods))

    @staticmethod
    def find_pod(name_prefix, namespace) -> "RobustaPod":
        pods: PodList = PodList.listNamespacedPod(namespace).obj
        for pod in pods.items:
            if pod.metadata.name.startswith(name_prefix):
                # we serialize and then deserialize to work around https://github.com/haxsaw/hikaru/issues/15
                return hikaru.from_dict(pod.to_dict(), cls=RobustaPod)
        raise Exception(
            f"No pod exists in namespace '{namespace}' with name prefix '{name_prefix}'"
        )

    # TODO: replace with Hikaru Pod().read() but note that usage is slightly different as this is a staticmethod
    @staticmethod
    def read(name: str, namespace: str) -> "RobustaPod":
        """Read pod definition from the API server"""
        return Pod.readNamespacedPod(name, namespace).obj


class RobustaDeployment(Deployment):
    @classmethod
    def from_image(cls: Type[T], name, image="busybox", cmd=None) -> T:
        obj: RobustaDeployment = hikaru.from_dict(
            yaml.safe_load(get_deployment_yaml(name, image)), RobustaDeployment
        )
        obj.spec.template.spec.containers[0].command = prepare_pod_command(cmd)
        return obj

    def get_images(self) -> Dict[str, str]:
        return get_images(self.spec.template.spec.containers)


class RobustaJob(Job):
    def get_pods(self) -> List[RobustaPod]:
        """
        gets the pods associated with a job
        """
        pods: PodList = PodList.listNamespacedPod(
            self.metadata.namespace, label_selector=f"job-name = {self.metadata.name}"
        ).obj
        # we serialize and then deserialize to work around https://github.com/haxsaw/hikaru/issues/15
        return [hikaru.from_dict(pod.to_dict(), cls=RobustaPod) for pod in pods.items]

    def get_single_pod(self) -> RobustaPod:
        """
        like get_pods() but verifies that only one pod is associated with the job and returns that pod
        """
        pods = self.get_pods()
        if len(pods) != 1:
            raise Exception(f"got more pods than expected for job: {pods}")
        return pods[0]

    @classmethod
    def run_simple_job_spec(cls, spec, name, timeout) -> str:
        job = RobustaJob(
            metadata=ObjectMeta(
                namespace=INSTALLATION_NAMESPACE, name=to_kubernetes_name(name)
            ),
            spec=JobSpec(
                backoffLimit=0,
                template=PodTemplateSpec(
                    spec=spec,
                ),
            ),
        )
        try:
            job = job.createNamespacedJob(job.metadata.namespace).obj
            job = hikaru.from_dict(
                job.to_dict(), cls=RobustaJob
            )  # temporary workaround for hikaru bug #15
            job: RobustaJob = wait_until_job_complete(job, timeout)
            job = hikaru.from_dict(
                job.to_dict(), cls=RobustaJob
            )  # temporary workaround for hikaru bug #15
            pod = job.get_single_pod()
            return pod.get_logs()
        finally:
            job.deleteNamespacedJob(
                job.metadata.name,
                job.metadata.namespace,
                propagation_policy="Foreground",
            )

    @classmethod
    def run_simple_job(cls, image, command, timeout) -> str:
        spec = PodSpec(
            containers=[
                Container(
                    name=to_kubernetes_name(image),
                    image=image,
                    command=prepare_pod_command(command),
                )
            ],
            restartPolicy="Never",
        )
        return cls.run_simple_job_spec(spec, name=image, timeout=timeout)


hikaru.register_version_kind_class(RobustaPod, Pod.apiVersion, Pod.kind)
hikaru.register_version_kind_class(
    RobustaDeployment, Deployment.apiVersion, Deployment.kind
)
hikaru.register_version_kind_class(RobustaJob, Job.apiVersion, Job.kind)



# This file was autogenerated. Do not edit.

from .v1.models import KIND_TO_MODEL_CLASS as v1
from .v2beta1.models import KIND_TO_MODEL_CLASS as v2beta1
from .v2beta2.models import KIND_TO_MODEL_CLASS as v2beta2

VERSION_KIND_TO_MODEL_CLASS = {"v1": v1, "v2beta1": v2beta1, "v2beta2": v2beta2}


def get_api_version(apiVersion: str):
    if "/" in apiVersion:
        apiVersion = apiVersion.split("/")[1]
    return VERSION_KIND_TO_MODEL_CLASS.get(apiVersion)



import logging
import os
import os.path
from inspect import getmembers

import colorlog
import manhole
from flask import Flask, request

from ..core.model.env_vars import NUM_EVENT_THREADS
from ..utils.task_queue import TaskQueue
from ..core.model.cloud_event import CloudEvent
from .. import api as robusta_api
from ..core.active_playbooks import run_playbooks
from ..integrations.prometheus.incoming_handler import parse_incoming_prometheus_alerts
from ..integrations.manual.incoming_handler import parse_incoming_manual_trigger
from ..integrations.slack.receiver import start_slack_receiver
from .config_loader import ConfigLoader
from ..core.schedule.scheduler import Scheduler


app = Flask(__name__)

LOGGING_LEVEL = os.environ.get("LOG_LEVEL", "INFO")
LOGGING_FORMAT = "%(log_color)s%(asctime)s.%(msecs)03d %(levelname)-8s %(message)s"
LOGGING_DATEFMT = "%Y-%m-%d %H:%M:%S"

if os.environ.get("ENABLE_COLORED_LOGS", "false").lower() == "true":
    print("setting up colored logging")
    colorlog.basicConfig(
        format=LOGGING_FORMAT, level=LOGGING_LEVEL, datefmt=LOGGING_DATEFMT
    )
else:
    print("setting up regular logging")
    logging.basicConfig(
        format=LOGGING_FORMAT, level=LOGGING_LEVEL, datefmt=LOGGING_DATEFMT
    )

logging.getLogger().setLevel(LOGGING_LEVEL)
log = logging.getLogger("werkzeug")
log.setLevel(logging.ERROR)
logging.info(f"logger initialized using {LOGGING_LEVEL} log level")
task_queue = TaskQueue(num_workers=NUM_EVENT_THREADS)


def main():
    Scheduler.init_scheduler()
    playbooks_path = os.environ.get("PLAYBOOKS_ROOT")
    config_file_path = os.environ.get("PLAYBOOKS_CONFIG_FILE_PATH")
    loader = ConfigLoader(config_file_path, playbooks_path)
    if os.environ.get("ENABLE_MANHOLE", "false").lower() == "true":
        manhole.install(locals=dict(getmembers(robusta_api)))

    start_slack_receiver()
    app.run(host="0.0.0.0", use_reloader=False)
    loader.close()


# TODO: in each of the below handlers we block until the playbook finishes running
# this is probably wrong especially if a playbook runs for some time
@app.route("/api/alerts", methods=["POST"])
def handle_alert_event():
    cloud_event = parse_incoming_prometheus_alerts(request)
    task_queue.add_task(run_playbooks, cloud_event)
    return "OK"


@app.route("/api/handle", methods=["POST"])
def handle_cloud_event():
    cloud_event = CloudEvent(**request.get_json())
    task_queue.add_task(run_playbooks, cloud_event)
    return "OK"


@app.route("/api/trigger", methods=["POST"])
def handle_manual_trigger():
    cloud_event = parse_incoming_manual_trigger(request)
    run_playbooks(cloud_event)
    return "OK"


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['robusta',
 'robusta.api',
 'robusta.cli',
 'robusta.core',
 'robusta.core.discovery',
 'robusta.core.model',
 'robusta.core.persistency',
 'robusta.core.playbooks',
 'robusta.core.playbooks.internal',
 'robusta.core.reporting',
 'robusta.core.schedule',
 'robusta.core.sinks',
 'robusta.core.sinks.datadog',
 'robusta.core.sinks.kafka',
 'robusta.core.sinks.robusta',
 'robusta.core.sinks.robusta.dal',
 'robusta.core.sinks.slack',
 'robusta.integrations',
 'robusta.integrations.git',
 'robusta.integrations.kubernetes',
 'robusta.integrations.kubernetes.autogenerated',
 'robusta.integrations.kubernetes.autogenerated.v1',
 'robusta.integrations.kubernetes.autogenerated.v2beta1',
 'robusta.integrations.kubernetes.autogenerated.v2beta2',
 'robusta.integrations.manual',
 'robusta.integrations.prometheus',
 'robusta.integrations.resource_analysis',
 'robusta.integrations.scheduled',
 'robusta.integrations.slack',
 'robusta.runner',
 'robusta.utils']

package_data = \
{'': ['*']}

install_requires = \
['CairoSVG>=2.5.2,<3.0.0',
 'Flask>=1.1.2,<2.0.0',
 'better-exceptions>=0.3.3,<0.4.0',
 'click-spinner>=0.1.10,<0.2.0',
 'colorlog>=5.0.1,<6.0.0',
 'datadog-api-client>=1.2.0,<2.0.0',
 'dulwich>=0.20.23,<0.21.0',
 'grafana-api>=1.0.3,<2.0.0',
 'hikaru>=0.5.1-beta.0,<0.6.0',
 'kafka-python>=2.0.2,<3.0.0',
 'kubernetes>=12.0.1,<13.0.0',
 'manhole>=1.8.0,<2.0.0',
 'prometheus-api-client>=0.4.2,<0.5.0',
 'pydantic>=1.8.1,<2.0.0',
 'slack-sdk>=3.7.0,<4.0.0',
 'supabase-py>=0.0.2,<0.0.3',
 'tabulate>=0.8.9,<0.9.0',
 'typer>=0.3.2,<0.4.0',
 'watchdog>=2.1.0,<3.0.0']

entry_points = \
{'console_scripts': ['robusta = robusta.cli.main:app']}

setup_kwargs = {
    'name': 'robusta-cli',
    'version': '0.6.8',
    'description': '',
    'long_description': None,
    'author': 'Natan Yellin',
    'author_email': 'aantn@users.noreply.github.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7.1,<4.0.0',
}


setup(**setup_kwargs)



from enum import Enum
from typing import Optional

from dataclasses import dataclass
from hikaru.meta import HikaruDocumentBase

from ...core.model.k8s_operation_type import K8sOperationType
from ...core.model.events import BaseEvent


@dataclass
class K8sBaseEvent(BaseEvent):
    description: str = ""
    operation: Optional[
        K8sOperationType
    ] = None  # because this dataclass needs to have defaults :(
    obj: Optional[
        HikaruDocumentBase
    ] = None  # marked as optional because this dataclass needs to have defaults :(
    old_obj: Optional[HikaruDocumentBase] = None  # same above



from pydantic import BaseModel


class ServiceInfo(BaseModel):
    name: str
    service_type: str
    namespace: str
    classification: str = "None"
    deleted: bool = False

    def get_service_key(self) -> str:
        return f"{self.namespace}/{self.service_type}/{self.name}"

    def __eq__(self, other):
        if not isinstance(other, ServiceInfo):
            return NotImplemented

        return (
            self.name == other.name
            and self.service_type == other.service_type
            and self.namespace == other.namespace
            and self.classification == other.classification
            and self.deleted == other.deleted
        )



from datetime import datetime
from pydantic import BaseModel
from typing import Dict, Any

# for deserializing incoming events in the cloudevent format
class CloudEvent(BaseModel):
    specversion: str
    type: str
    source: str
    subject: str
    id: str
    time: datetime
    datacontenttype: str
    data: Dict[Any, Any]



import logging
from kubernetes import client
from kubernetes.client import V1ServiceList
from kubernetes.client.models.v1_service import V1Service


def find_service_url(label_selector):
    """
    Get the url of an in-cluster service with a specific label
    """
    # we do it this way because there is a weird issue with hikaru's ServiceList.listServiceForAllNamespaces()
    v1 = client.CoreV1Api()
    svc_list: V1ServiceList = v1.list_service_for_all_namespaces(
        label_selector=label_selector
    )
    if not svc_list.items:
        return None
    svc: V1Service = svc_list.items[0]
    name = svc.metadata.name
    namespace = svc.metadata.namespace
    port = svc.spec.ports[0].port
    url = f"http://{name}.{namespace}.svc.cluster.local:{port}"
    logging.info(
        f"discovered service with label-selector: `{label_selector}` at url: `{url}`"
    )
    return url



import threading
import logging
from collections import defaultdict
from datetime import datetime


class RateLimiter:

    limiter_lock = threading.Lock()
    limiter_map = defaultdict(None)

    @staticmethod
    def mark_and_test(operation: str, id: str, period_seconds: int) -> bool:
        with RateLimiter.limiter_lock:
            limiter_key = operation + id
            last_run = RateLimiter.limiter_map.get(limiter_key)
            curr_seconds = datetime.utcnow().timestamp()
            if last_run:
                if curr_seconds - last_run > period_seconds:
                    RateLimiter.limiter_map[limiter_key] = curr_seconds
                    logging.info(
                        f"rate limited operation is allowed because enough time has passed: {operation}"
                    )
                    return True
                else:
                    logging.info(f"rate limited operation is NOT allowed: {operation}")
                    return False
            else:
                logging.info(
                    f"rate limited operation is allowed because it is the first time: {operation}"
                )
                RateLimiter.limiter_map[limiter_key] = curr_seconds
                return True



import subprocess
import time
from contextlib import contextmanager

import click_spinner
import typer
import requests

from robusta._version import __version__

PLAYBOOKS_DIR = "playbooks/"


def exec_in_robusta_runner(
    cmd,
    namespace: str,
    tries=1,
    time_between_attempts=10,
    error_msg="error running cmd",
):
    cmd = [
        "kubectl",
        "exec",
        "-n",
        namespace,
        "-it",
        "deploy/robusta-runner",
        "-c",
        "runner",
        "--",
        "bash",
        "-c",
        cmd,
    ]
    for _ in range(tries - 1):
        try:
            return subprocess.check_call(cmd)
        except Exception as e:
            typer.echo(f"{error_msg}")
            time.sleep(time_between_attempts)
    return subprocess.check_call(cmd)


def download_file(url, local_path):
    with click_spinner.spinner():
        response = requests.get(url)
        response.raise_for_status()
    with open(local_path, "wb") as f:
        f.write(response.content)


def log_title(title, color=None):
    typer.echo("=" * 70)
    typer.secho(title, fg=color)
    typer.echo("=" * 70)


def replace_in_file(path, original, replacement):
    with open(path) as r:
        text = r.read()
        if original not in text:
            raise Exception(
                f"Cannot replace text {original} in file {path} because it was not found"
            )
        text = text.replace(original, replacement)
    with open(path, "w") as w:
        w.write(text)


@contextmanager
def fetch_runner_logs(namespace: str, all_logs=False):
    start = time.time()
    try:
        yield
    finally:
        log_title("Fetching logs...")
        if all_logs:
            subprocess.check_call(
                f"kubectl logs -n {namespace} deployment/robusta-runner -c runner",
                shell=True,
            )
        else:
            subprocess.check_call(
                f"kubectl logs -n {namespace} deployment/robusta-runner -c runner --since={int(time.time() - start + 1)}s",
                shell=True,
            )


def get_examples_url(examples_version=None):
    if examples_version is None:
        examples_version = __version__
    return f"https://storage.googleapis.com/robusta-public/{examples_version}/example-playbooks.zip"



import json
import logging
import time

from pydantic import BaseModel
from datadog_api_client.v1 import ApiClient, ApiException, Configuration
from datadog_api_client.v1.api import events_api
from datadog_api_client.v1.models import *

from typing import List
from tabulate import tabulate

from ..sink_config import SinkConfigBase
from ...reporting.blocks import (
    Finding,
    Enrichment,
    KubernetesDiffBlock,
    JsonBlock,
    FindingSeverity,
    MarkdownBlock,
    FileBlock,
    HeaderBlock,
    ListBlock,
    TableBlock,
    DividerBlock,
)
from ..sink_base import SinkBase


class DataDogSinkConfig(BaseModel):
    api_key: str


class DataDogSink(SinkBase):
    def __init__(self, sink_config: SinkConfigBase, cluster_name: str):
        super().__init__(sink_config)
        self.cluster_name = cluster_name
        config = Configuration()
        config.api_key["apiKeyAuth"] = DataDogSinkConfig(**sink_config.params).api_key
        self.api_instance = events_api.EventsApi(ApiClient(config))

    @staticmethod
    def __to_datadog_event_type(severity: FindingSeverity):
        # must be one of ["error", "warning", "info", "success", "user_update", "recommendation", "snapshot", ]
        if severity == FindingSeverity.HIGH:
            return EventAlertType("error")
        elif severity == FindingSeverity.MEDIUM:
            return EventAlertType("warning")
        else:
            return EventAlertType("info")

    @staticmethod
    def __enrichment_text(enrichment: Enrichment) -> str:
        lines = []
        for block in enrichment.blocks:
            if isinstance(block, MarkdownBlock):
                if not block.text:
                    continue
                lines.append(f"%%%\n{block.text}\n%%%")
            elif isinstance(block, DividerBlock):
                lines.append("-------------------")
            elif isinstance(block, JsonBlock):
                lines.append(block.json_str)
            elif isinstance(block, KubernetesDiffBlock):
                for diff in block.diffs:
                    lines.append(
                        f"*{'.'.join(diff.path)}*: {diff.other_value} ==> {diff.value}"
                    )
            elif isinstance(block, FileBlock):
                last_dot_idx = block.filename.rindex(".")
                file_type = block.filename[last_dot_idx + 1 :]
                if file_type == "txt":
                    lines.append(block.filename)
                    lines.append("------------------")
                    lines.extend(str(block.contents).split("\n"))
                    lines.append("------------------")
            elif isinstance(block, HeaderBlock):
                lines.append(block.text)
                lines.append("------------------")
            elif isinstance(block, ListBlock):
                lines.extend(block.items)
            elif isinstance(block, TableBlock):
                rendered_rows = block.render_rows()
                lines.append(
                    tabulate(rendered_rows, headers=block.headers, tablefmt="presto")
                )
        return "\n".join(lines)

    @staticmethod
    def __enrichments_as_text(enrichments: List[Enrichment]) -> str:
        text_arr = [
            DataDogSink.__enrichment_text(enrichment) for enrichment in enrichments
        ]
        return "---\n".join(text_arr)

    @staticmethod
    def __trim_str(text: str, size_limit: int) -> str:
        if len(text) >= size_limit:
            text = text[:size_limit] + "..."
        return text

    def write_finding(self, finding: Finding):
        resource = finding.subject
        body = EventCreateRequest(
            aggregation_key=finding.aggregation_key,
            alert_type=DataDogSink.__to_datadog_event_type(finding.severity),
            date_happened=int(time.time()),
            source_type_name="Robusta",
            host=f"{resource.namespace}/{resource.subject_type.value}/{resource.name}",
            tags=[f"cluster:{self.cluster_name}"],
            text=DataDogSink.__enrichments_as_text(finding.enrichments),
            title=DataDogSink.__trim_str(finding.title, 97),
        )

        try:
            api_response = self.api_instance.create_event(body)
            logging.debug(f"datadog event api response {api_response}")
        except ApiException as e:
            logging.error(
                "Exception when calling datadog EventsApi->create_event: %s\n" % e
            )



# dummy persistence driver
from contextlib import contextmanager
from pydantic import BaseModel
from typing import Type, TypeVar, Dict, ContextManager

persistent_data: Dict[str, BaseModel] = {}

# TODO: we probably want some form of locking for this so two playbooks can't edit the same data at the same time
T = TypeVar("T", bound=BaseModel)


@contextmanager
def get_persistent_data(name: str, cls: Type[T]) -> ContextManager[T]:
    try:
        data = persistent_data.get(name, cls())
        yield data
    finally:
        # write data back
        persistent_data[name] = data



import logging
import os
import shutil
import textwrap
import threading
from collections import defaultdict
import traceback
import uuid
from datetime import datetime

from dulwich import porcelain

GIT_DIR_NAME = "robusta-git"
REPO_LOCAL_BASE_DIR = os.path.join(
    os.environ.get("REPO_LOCAL_BASE_DIR", "/app"), GIT_DIR_NAME
)
SSH_ROOT_DIR = os.environ.get("SSH_ROOT_DIR", "/root/.ssh")


class GitRepoManager:

    manager_lock = threading.Lock()
    repo_map = defaultdict(None)

    @staticmethod
    def get_git_repo(git_repo_url: str, git_key: str):
        with GitRepoManager.manager_lock:
            repo = GitRepoManager.repo_map.get(git_repo_url)
            if repo is not None:
                return repo
            repo = GitRepo(git_repo_url, git_key)
            GitRepoManager.repo_map[git_repo_url] = repo
            return repo

    @staticmethod
    def remove_git_repo(git_repo_url):
        with GitRepoManager.manager_lock:
            del GitRepoManager.repo_map[git_repo_url]

    @staticmethod
    def clear_git_repos():
        with GitRepoManager.manager_lock:
            GitRepoManager.repo_map.clear()


class GitRepo:

    initialized: bool = False

    def __init__(self, git_repo_url: str, git_key: str):
        GitRepo.init()
        self.key_file_name = self.init_key(git_key)
        self.repo_lock = threading.RLock()
        self.git_repo_url = git_repo_url
        self.repo_name = os.path.splitext(os.path.basename(git_repo_url))[0]
        self.repo_local_path = os.path.join(REPO_LOCAL_BASE_DIR, self.repo_name)
        self.init_repo()

    def init_key(self, git_key):
        pkey_name = str(uuid.uuid4())
        key_file_name = os.path.join(REPO_LOCAL_BASE_DIR, pkey_name)
        with open(key_file_name, "w") as key_file:
            key_file.write(textwrap.dedent(f"{git_key}"))
        os.chmod(key_file_name, 0o400)
        if not os.path.exists(SSH_ROOT_DIR):
            os.mkdir(SSH_ROOT_DIR)
        os.system(f"ssh-keyscan -H github.com >> {SSH_ROOT_DIR}/known_hosts")
        return key_file_name

    @staticmethod
    def init():
        if GitRepo.initialized:
            return
        try:
            os.makedirs(REPO_LOCAL_BASE_DIR, exist_ok=True)
        except Exception as e:
            print(
                f"Failed to create git audit base path {REPO_LOCAL_BASE_DIR}",
                traceback.print_exc(),
            )
            raise e
        GitRepo.initialized = True

    def init_repo(self):
        with self.repo_lock:
            if os.path.exists(self.repo_local_path):
                logging.info(f"Deleting local repo before init {self.repo_local_path}")
                shutil.rmtree(self.repo_local_path)

            logging.info(
                f"Cloning git repo {self.git_repo_url}. repo name {self.repo_name}"
            )
            self.repo = porcelain.clone(
                self.git_repo_url, self.repo_local_path, key_filename=self.key_file_name
            )

    def commit(
        self,
        file_data: str,
        file_path: str,
        file_name,
        commit_message: str = "Robusta Git",
    ):
        with self.repo_lock:
            self.pull_rebase()
            file_local_path = os.path.join(self.repo_local_path, file_path)
            try:
                os.makedirs(file_local_path, exist_ok=True)
                git_file_name = os.path.join(file_local_path, file_name)
                with open(git_file_name, "w") as git_file:
                    git_file.write(file_data)

                porcelain.add(self.repo, git_file_name)
                porcelain.commit(self.repo, commit_message)
            except Exception as e:
                logging.error(
                    f"Commit file failed {self.repo_local_path} {file_path} {file_name}",
                    traceback.print_exc(),
                )
                GitRepoManager.remove_git_repo(self.git_repo_url)
                raise e

    def push(self):
        with self.repo_lock:
            try:
                porcelain.push(self.repo, key_filename=self.key_file_name)
            except Exception as e:
                GitRepoManager.remove_git_repo(self.git_repo_url)
                logging.error(
                    f"Push failed {self.repo_local_path}", traceback.print_exc()
                )
                raise e

    def pull_rebase(self):
        with self.repo_lock:
            os.system(
                f"cd {self.repo_local_path} && GIT_SSH_COMMAND='ssh -i {self.key_file_name} -o IdentitiesOnly=yes' git pull --rebase origin master"
            )

    def commit_push(
        self, file_data: str, file_path: str, file_name, commit_message: str
    ):
        with self.repo_lock:
            self.commit(file_data, file_path, file_name, commit_message)
            self.push()

    def delete(self, file_path: str, file_name):
        with self.repo_lock:
            file_local_path = os.path.join(self.repo_local_path, file_path)
            if not os.path.exists(
                file_local_path
            ):  # file not in repo. Might have been added before the audit playbook was configured
                return

            try:
                self.pull_rebase()
                porcelain.remove(self.repo, [os.path.join(file_local_path, file_name)])
                porcelain.commit(self.repo, f"robusta audit {datetime.now()} - delete")
            except Exception as e:
                logging.error(
                    f"Commit file failed {self.repo_local_path} {file_path} {file_name}",
                    traceback.print_exc(),
                )
                GitRepoManager.remove_git_repo(self.git_repo_url)
                raise e

    def delete_push(self, file_path: str, file_name):
        with self.repo_lock:
            self.delete(file_path, file_name)
            self.push()



import copy
import logging
import traceback

from typing import List

from ..core.sinks.sink_manager import SinkManager
from ..core.model.events import BaseEvent


def handle_event(
    func, event: BaseEvent, action_params, event_source: str, named_sinks: List[str]
):
    event_desc = getattr(event, "description", None)
    description = f"event={event_desc}" if event_desc else ""
    logging.info(
        f"running {event_source} playbook {func.__name__}; action_params={action_params}; sinks={named_sinks}; {description}"
    )
    event.named_sinks = named_sinks
    if action_params is None:
        result = func(event)
    else:
        result = func(event, action_params)

    if event.finding:
        for sink_name in named_sinks:
            try:
                sink = SinkManager.get_sink_by_name(sink_name)
                if not sink:
                    logging.error(
                        f"sink {sink_name} not found. Skipping event finding {event.finding}"
                    )
                    continue
                # create deep copy, so that iterating on one sink won't affect the others
                finding_copy = copy.deepcopy(event.finding)
                sink.write_finding(finding_copy)
            except Exception:  # Faliure to send to one sink shouldn't fail all
                logging.error(
                    f"Failed to publish finding to sink {sink_name}",
                    traceback.print_exc(),
                )

    if result is not None:
        return result
    return "OK"
